function A = alpertizematrix(A, partnoa, kerfun, o) 
% A = alpertizeselfmatrix(A,,,.... **** returns Alpert quadrature rule
%  matrix corrected along a band around the diagonal, given the matrix A of all
%  source-to-target kernel values (including speed function and uniform
%  quadrature weights 2pi/N), with diag(A)=0. kerfun is a
%  kernel function of the form
%     val = kerfun(targs, src, srcnormal)
%  returning k(s,t) without the speed |z'(t)| factor.
%  s is the segment whose self-interaction is needed.
%  opts.ord controls Alpert quadrature order; other options ignored.
%  Now vectorized for speed! (Most time is spent on kernel evaluation, good)
%
%  format for [x speed nor] = partnoa(t)
%
%  If segment.qpblocha is nonempty, segment is treated as grating periodic with
%   Bloch phase qpblocha, and segment parametrization is assumed to spill over
%   correctly into t<0 and t>1. Fixes are done appropriate for nei>0 QP scheme.

% 2011 Alex Barnett, edited 2012 Brad Nelson.
M = size(A,1); N = size(A,2);
if M~=N, warning('Alpert quadr will fail unless matrix square!'); end

[tex,wex,nskip] = QuadLogExtraPtNodes(o.ord); % get log nodes, wei
tex = [-tex(end:-1:1); tex]; wex = [wex(end:-1:1); wex]; % make symmetric
wex = wex/N;      % scale for curve regular node parameter spacing
Na = numel(tex);   % # Alpert pts (both sides)
if N<=2*nskip, warning('not enough regular nodes for Alpert order!'); end
ninterp = o.ord + 3;   % # local regular Lagrange interpolation pts (O'Neil=+2)

% For Alpert, first need to kill off the band-diagonal of A:
% (following loops are over rows, i=targets, j=source density dofs).
% s closed: kill wrapped band diagonal (ie including SW and NE corners)...
  for i=1:N, A(i, mod(i+[-nskip+1:nskip-1]-1, N)+1) = 0; end

%return   % for debug!

% Do set-up for Alpert quadrature extra nodes and weights...
% These arrays are N-by-Na... note that t params can spill over from [0,1]
t = (0:(N-1))/N*2*pi;   % same as in param, row vector
[x speed n] = partnoa(t);
x = repmat(x, [1 Na]); % 2 by N*Na, Na is slow varying
%nx = repmat(n, [1 Na]); % target locations & normals  , not needed
tt = repmat(t, [1 Na]) + kron(2*pi*tex.'/N, ones(1,N)); % src curve params
%figure; plot(tt); drawnow;
tt = mod(tt,2*pi);                         % wrap into [0,1]
% Get all curve info from segment at once = fast! 
[y sp ny] = partnoa(tt);  % sp is speeds at src pts y
%figure; plot(y(1,1:N:end), y(2,1:N:end), '.'); drawnow;

%K = zeros(N, Na);
%for m=1:N, for l=1:Na, K(m,l) = sp(m+(l-1)*N) * kerfun(x(:,
xc=x(1,:)'+1i*(x(2,:)');
yc=y(1,:)'+1i*(y(2,:)');
nyc=ny(1,:)'+1i*(ny(2,:)');
K = sp .* kerfun(xc, yc, nyc).'; % get all kernel evals at once = fast
K = reshape(K, [N Na]);
K=K*2*pi;
%figure; imagesc(imag(K)); colorbar; drawnow;

for l=1:numel(tex)  % loop over Alpert's nonregular quadr pts
  dt = tex(l)/N;    % parameter offset for current quadr pt
  joffs = floor(tex(l)-ninterp/2+1) + (0:ninterp-1); % nearest interp pts
  % interpolation pts and L = row vec of Lagrange basis funcs eval at dt:
  x = joffs/N; w = baryweights(x); L = baryprojs(x, w, dt);
  ii = repmat(1:N, [1 ninterp]);  % i-indices of A to write over
  jj = ii + kron(joffs, ones(1,N));  % j-indices not yet wrapped into [1,N]

  jj = mod(jj-1, N) + 1;             % j-indices of A to write over
  iii = sub2ind(size(A), ii,jj);     % 1d array indices ditto
  A(iii) = A(iii) + wex(l) * repmat(K(:,l).', [1 ninterp]) .* kron(L,ones(1,N));  % This is probably slow because accesses a lot of A's memory for each l!
end
% Note the affected elements spill out to dist nskip-1+iord/2 from diag!
% (even though the diag band that was set to zero is only to dist nskip-1)
